from itertools import product
from pathlib import Path

import asreview as asr
import pytest
from asreview.extensions import extensions
from asreview.models.balancers import Balanced
from asreview.models.queriers import Max

classifier_parameters = {
    "xgboost": {"max_depth": 5, "n_estimators": 250},
    "dynamic-nn": {"epochs": 30, "batch_size": 16},
    "nn-2-layer": {"epochs": 50, "verbose": 1},
    "warmstart-nn": {"epochs": 45, "shuffle": False},
    "adaboost": {"n_estimators": 30, "learning_rate": 0.5},
}

feature_extractor_parameters = {
    "labse": {"normalize": "l2", "quantize": False},
    "mxbai": {
        "normalize": "minmax",
        "precision": "binary",
        "quantize": True,
    },
    "sbert": {
        "normalize": "standard",
        "verbose": False,
        "quantize": False,
    },
    "multilingual-e5-large": {"normalize": False, "sep": ",", "quantize": True},
    "gtr-t5-large": {"normalize": True, "columns": ["title"], "quantize": False},
}

# Get all classifiers and feature extractors from ASReview, filtering contrib models
classifiers = [
    cls for cls in extensions("models.classifiers") if "asreviewcontrib" in str(cls)
]
feature_extractors = [
    fe for fe in extensions("models.feature_extractors") if "asreviewcontrib" in str(fe)
]

# Generate all combinations of classifier and feature extractor
pairs = list(product(classifiers, feature_extractors))

test_ids = [
    f"{classifier.name}__{feature_extractor.name}"
    for classifier, feature_extractor in pairs
]


@pytest.mark.parametrize("classifier, feature_extractor", pairs, ids=test_ids)
def test_alc_to_and_from_meta(classifier, feature_extractor):
    alc1 = asr.ActiveLearningCycle(
        classifier=classifier.load()(**classifier_parameters.get(classifier.name)),
        feature_extractor=feature_extractor.load()(
            **feature_extractor_parameters.get(feature_extractor.name)
        ),
        balancer=Balanced(ratio=5),
        querier=Max(),
    )

    alc2_meta = asr.ActiveLearningCycleData(
        classifier=classifier.name,
        classifier_param=classifier_parameters.get(classifier.name),
        feature_extractor=feature_extractor.name,
        feature_extractor_param=feature_extractor_parameters.get(
            feature_extractor.name
        ),
        balancer="balanced",
        balancer_param={"ratio": 5},
        querier="max",
    )

    alc1_meta = alc1.to_meta()

    alc1_from_meta = asr.ActiveLearningCycle.from_meta(alc1_meta)

    alc2_from_meta = asr.ActiveLearningCycle.from_meta(alc2_meta)

    assert (
        alc1.classifier.name
        == alc1_from_meta.classifier.name
        == alc2_from_meta.classifier.name
    ), "Classifier names do not match"
    assert (
        alc1.classifier.get_params(deep=False)
        == alc1_from_meta.classifier.get_params(deep=False)
        == alc2_from_meta.classifier.get_params(deep=False)
    ), "Classifier parameters do not match"

    assert (
        alc1.feature_extractor.name
        == alc1_from_meta.feature_extractor.name
        == alc2_from_meta.feature_extractor.name
    ), "Feature extractor names do not match"
    assert (
        alc1.feature_extractor.get_params(deep=False)
        == alc1_from_meta.feature_extractor.get_params(deep=False)
        == alc2_from_meta.feature_extractor.get_params(deep=False)
    ), "Feature extractor parameters do not match"

    assert (
        alc1.balancer.name
        == alc1_from_meta.balancer.name
        == alc2_from_meta.balancer.name
    ), "Balancer names do not match"
    assert (
        alc1.balancer.get_params(deep=False)
        == alc1_from_meta.balancer.get_params(deep=False)
        == alc2_from_meta.balancer.get_params(deep=False)
    ), "Balancer parameters do not match"

    assert (
        alc1.querier.name == alc1_from_meta.querier.name == alc2_from_meta.querier.name
    ), "Querier names do not match"
    assert (
        alc1.querier.get_params(deep=False)
        == alc1_from_meta.querier.get_params(deep=False)
        == alc2_from_meta.querier.get_params(deep=False)
    ), "Querier parameters do not match"


@pytest.mark.parametrize("classifier, feature_extractor", pairs, ids=test_ids)
def test_alc_to_and_from_file(tmpdir, classifier, feature_extractor):
    alc1 = asr.ActiveLearningCycle(
        classifier=classifier.load()(**classifier_parameters.get(classifier.name)),
        feature_extractor=feature_extractor.load()(
            **feature_extractor_parameters.get(feature_extractor.name)
        ),
        balancer=Balanced(ratio=5),
        querier=Max(),
    )

    alc2_meta = asr.ActiveLearningCycleData(
        classifier=classifier.name,
        classifier_param=classifier_parameters.get(classifier.name),
        feature_extractor=feature_extractor.name,
        feature_extractor_param=feature_extractor_parameters.get(
            feature_extractor.name
        ),
        balancer="balanced",
        balancer_param={"ratio": 5},
        querier="max",
    )

    meta_file_path = Path(tmpdir, "alc1.json")
    alc1.to_file(meta_file_path)
    alc1_from_file = asr.ActiveLearningCycle.from_file(meta_file_path)

    alc2_from_meta = asr.ActiveLearningCycle.from_meta(alc2_meta)

    assert (
        alc1.classifier.name
        == alc1_from_file.classifier.name
        == alc2_from_meta.classifier.name
    ), "Classifier names do not match"

    assert (
        alc1.classifier.get_params(deep=False)
        == alc1.classifier.get_params(deep=False)
        == alc1.classifier.get_params(deep=False)
    ), "Classifier parameters do not match"

    assert (
        alc1.feature_extractor.name
        == alc1_from_file.feature_extractor.name
        == alc2_from_meta.feature_extractor.name
    ), "Feature extractor names do not match"
    assert (
        alc1.feature_extractor.get_params(deep=False)
        == alc1_from_file.feature_extractor.get_params(deep=False)
        == alc2_from_meta.feature_extractor.get_params(deep=False)
    ), "Feature extractor parameters do not match"

    assert (
        alc1.balancer.name
        == alc1_from_file.balancer.name
        == alc2_from_meta.balancer.name
    ), "Balancer names do not match"
    assert (
        alc1.balancer.get_params(deep=False)
        == alc1_from_file.balancer.get_params(deep=False)
        == alc2_from_meta.balancer.get_params(deep=False)
    ), "Balancer parameters do not match"

    assert (
        alc1.querier.name == alc1_from_file.querier.name == alc2_from_meta.querier.name
    ), "Querier names do not match"
    assert (
        alc1.querier.get_params(deep=False)
        == alc1_from_file.querier.get_params(deep=False)
        == alc2_from_meta.querier.get_params(deep=False)
    ), "Querier parameters do not match"



__all__ = ["DynamicNNClassifier", "NN2LayerClassifier", "WarmStartNNClassifier"]

import os
from math import ceil, log10

os.environ["KERAS_BACKEND"] = "torch"

from keras import layers, losses, models, optimizers, regularizers, wrappers


class BaseNNClassifier(wrappers.SKLearnClassifier):
    """
    Base Neural Network Classifier.

    Subclasses must implement the _build_nn_model method.
    """

    def __init__(self, **kwargs):
        fit_kwargs = {
            "epochs": kwargs.pop("epochs", 35),
            "verbose": kwargs.pop("verbose", 0),
            "batch_size": kwargs.pop("batch_size", 32),
            "shuffle": kwargs.pop("shuffle", True),
        }

        super().__init__(
            model=self._build_nn_model, model_kwargs=kwargs, fit_kwargs=fit_kwargs
        )

    def _build_nn_model(self, X, y):
        raise NotImplementedError(
            "Subclasses should implement the _build_nn_model method."
        )

    def predict_proba(self, X):
        return self.model_.predict(X, verbose=0)


class DynamicNNClassifier(BaseNNClassifier):
    """
    Dynamic Neural Network Classifier

    Fully connected neural network classifier that dynamically selects
    the number of dense layers based on dataset size.
    """

    name = "dynamic-nn"
    label = "Fully connected neural network (dynamic layer count)"

    def _build_nn_model(self, X, y):
        input_dim = X.shape[1]
        num_layers = min(4, ceil(log10(max(10, X.shape[0]))))

        model = models.Sequential()
        model.add(layers.Input(shape=(input_dim,)))
        model.add(layers.BatchNormalization())
        for _ in range(num_layers):
            model.add(
                layers.Dense(
                    64,
                    activation="relu",
                    kernel_regularizer=regularizers.L2(),
                    kernel_initializer="he_normal",
                )
            )
            model.add(layers.Dropout(0.5))
        model.add(
            layers.Dense(
                y.shape[1] if len(y.shape) > 1 else 1,
                activation="sigmoid",
            )
        )

        model.compile(
            optimizer=optimizers.Adam(),
            loss=losses.BinaryCrossentropy(),
            metrics=["accuracy"],
        )

        return model


class NN2LayerClassifier(BaseNNClassifier):
    """Fully connected neural network (2 hidden layers) classifier (``nn-2-layer``).

    Neural network with two hidden, dense layers of the same size.
    """

    name = "nn-2-layer"
    label = "Fully connected neural network (2 hidden layers)"

    def _build_nn_model(self, X, y):
        input_dim = X.shape[1]

        model = models.Sequential()

        model.add(layers.Input(shape=(input_dim,)))

        model.add(
            layers.Dense(
                128,
                kernel_regularizer=regularizers.l2(0.01),
                activity_regularizer=regularizers.l1(0.01),
                activation="relu",
            )
        )

        model.add(
            layers.Dense(
                128,
                kernel_regularizer=regularizers.l2(0.01),
                activity_regularizer=regularizers.l1(0.01),
                activation="relu",
            )
        )

        model.add(
            layers.Dense(y.shape[1] if len(y.shape) > 1 else 1, activation="sigmoid")
        )

        # Compile model
        model.compile(
            loss="binary_crossentropy",
            optimizer=optimizers.RMSprop(learning_rate=0.001),
            metrics=["acc"],
        )

        return model


class WarmStartNNClassifier(BaseNNClassifier):
    """
    Neural network with warm-starting behavior.

    Retains previous weights and uses them as initial state on subsequent fits.
    """

    name = "warmstart-nn"
    label = "Neural network (warm start, 2 hidden layers)"

    _last_weights = None

    def _build_nn_model(self, X, y):
        input_dim = X.shape[1]

        model = models.Sequential()
        model.add(layers.Input(shape=(input_dim,)))
        model.add(
            layers.Dense(
                128,
                kernel_regularizer=regularizers.l2(0.01),
                activity_regularizer=regularizers.l1(0.01),
                activation="relu",
            )
        )
        model.add(
            layers.Dense(
                128,
                kernel_regularizer=regularizers.l2(0.01),
                activity_regularizer=regularizers.l1(0.01),
                activation="relu",
            )
        )
        model.add(
            layers.Dense(
                y.shape[1] if len(y.shape) > 1 else 1,
                activation="sigmoid",
            )
        )

        model.compile(
            loss="binary_crossentropy",
            optimizer=optimizers.RMSprop(learning_rate=0.001),
            metrics=["acc"],
        )

        if WarmStartNNClassifier._last_weights is not None:
            try:
                model.set_weights(WarmStartNNClassifier._last_weights)
            except ValueError:
                # Handle cases where the model architecture has changed
                print(
                    "Warning: Previous weights do not match current model architecture."
                )
                WarmStartNNClassifier._last_weights = None
        return model

    def fit(self, X, y, **kwargs):
        super().fit(X, y, **kwargs)
        WarmStartNNClassifier._last_weights = self.model_.get_weights()
        return self



from pathlib import Path

import asreview as asr
import pandas as pd
import pytest
from asreview.extensions import extensions, get_extension
from asreview.models.balancers import Balanced
from asreview.models.queriers import Max

from asreviewcontrib.dory.entrypoint import DoryEntryPoint

classifier_parameters = {
    "xgboost": {"max_depth": 5, "n_estimators": 250},
    "dynamic-nn": {"epochs": 30, "batch_size": 16},
    "nn-2-layer": {"epochs": 50, "verbose": 1},
    "warmstart-nn": {"epochs": 45, "shuffle": False},
    "adaboost": {"n_estimators": 30, "learning_rate": 0.5},
}

feature_extractor_parameters = {
    "labse": {"normalize": "l2", "quantize": False},
    "mxbai": {
        "normalize": "minmax",
        "precision": "binary",
        "quantize": True,
    },
    "sbert": {
        "normalize": "standard",
        "verbose": False,
        "quantize": False,
    },
    "multilingual-e5-large": {"normalize": False, "sep": ",", "quantize": True},
    "gtr-t5-large": {"normalize": True, "columns": ["title"], "quantize": False},
}

# Define dataset path
dataset_path = Path("tests/data/generic_labels.csv")

# Get all classifiers and feature extractors from ASReview, filtering contrib models
classifiers = [
    cls for cls in extensions("models.classifiers") if "asreviewcontrib" in str(cls)
]
feature_extractors = [
    fe for fe in extensions("models.feature_extractors") if "asreviewcontrib" in str(fe)
]

test_ids = [
    f"{feature_extractor.name}__per_classifier"
    for feature_extractor in feature_extractors
]


@pytest.mark.parametrize("feature_extractor", feature_extractors, ids=test_ids)
def test_all_fe_clf_combinations(feature_extractor):
    # Load dataset
    data = asr.load_dataset(dataset_path)

    # Preprocess data using feature extractor
    fm = feature_extractor.load()(
        **feature_extractor_parameters.get(feature_extractor.name)
    ).fit_transform(data)

    # Test each classifier on the preprocessed FMs
    for classifier in classifiers:
        # Define Active Learning Cycle
        alc = asr.ActiveLearningCycle(
            classifier=classifier.load()(**classifier_parameters.get(classifier.name)),
            feature_extractor=feature_extractor.load()(
                **feature_extractor_parameters.get(feature_extractor.name)
            ),
            balancer=None,
            querier=Max(),
        )

        # Run simulation
        simulate = asr.Simulate(
            X=fm,
            labels=data["included"],
            cycles=[alc],
            skip_transform=True,
        )
        simulate.label([0, 1])
        simulate.review()

        assert isinstance(simulate._results, pd.DataFrame)
        assert simulate._results.shape[0] > 2 and simulate._results.shape[0] <= 6, (
            "Simulation produced incorrect number of results."
        )
        assert classifier.name in simulate._results["classifier"].unique(), (
            "Classifier is not in results."
        )
        assert (
            feature_extractor.name in simulate._results["feature_extractor"].unique()
        ), "Feature extractor is not in results."


def test_language_agnostic_l2_preset():
    # Load dataset
    data = asr.load_dataset(dataset_path)

    # Define Active Learning Cycle
    alc = asr.ActiveLearningCycle(
        classifier=get_extension("models.classifiers", "svm").load()(
            loss="squared_hinge", C=0.106, max_iter=5000
        ),
        feature_extractor=get_extension(
            "models.feature_extractors", "multilingual-e5-large"
        ).load()(normalize=True),
        balancer=Balanced(ratio=9.707),
        querier=Max(),
    )
    # Run simulation
    simulate = asr.Simulate(
        X=data,
        labels=data["included"],
        cycles=[alc],
    )
    simulate.label([0, 1])
    simulate.review()
    assert isinstance(simulate._results, pd.DataFrame)
    assert simulate._results.shape[0] > 2 and simulate._results.shape[0] <= 6, (
        "Simulation produced incorrect number of results."
    )
    assert (
        get_extension("models.classifiers", "svm").load()().name
        in simulate._results["classifier"].unique()
    ), "Classifier is not in results."
    assert (
        get_extension("models.feature_extractors", "multilingual-e5-large")
        .load()()
        .name
        in simulate._results["feature_extractor"].unique()
    ), "Feature extractor is not in results."


def test_heavy_h3_preset():
    # Load dataset
    data = asr.load_dataset(dataset_path)

    # Define Active Learning Cycle
    alc = asr.ActiveLearningCycle(
        classifier=get_extension("models.classifiers", "svm").load()(
            loss="squared_hinge", C=0.067, max_iter=5000
        ),
        feature_extractor=get_extension("models.feature_extractors", "mxbai").load()(
            normalize=True
        ),
        balancer=Balanced(ratio=9.724),
        querier=Max(),
    )
    # Run simulation
    simulate = asr.Simulate(
        X=data,
        labels=data["included"],
        cycles=[alc],
    )
    simulate.label([0, 1])
    simulate.review()
    assert isinstance(simulate._results, pd.DataFrame)
    assert simulate._results.shape[0] > 2 and simulate._results.shape[0] <= 6, (
        "Simulation produced incorrect number of results."
    )
    assert (
        get_extension("models.classifiers", "svm").load()().name
        in simulate._results["classifier"].unique()
    ), "Classifier is not in results."
    assert (
        get_extension("models.feature_extractors", "mxbai").load()().name
        in simulate._results["feature_extractor"].unique()
    ), "Feature extractor is not in results."


def test_get_all_models():
    assert len(DoryEntryPoint()._get_all_models()) == 10

def test_invalid_normalization_method():
    fe = get_extension("models.feature_extractors", "multilingual-e5-large").load()

    with pytest.raises(ValueError, match="Unsupported normalization method"):
        fe(normalize="invalid-method")
